#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.
import logging
import time
import threading
import amqpstorm
from amqpstorm import Connection
from amqpstorm import UriConnection
import json
import requests

logging.basicConfig(level=logging.CRITICAL)

LOGGER = logging.getLogger()

class _MQConsumer(threading.Thread):
    def __init__(self, connection):
        threading.Thread.__init__(self, None, None, "MQConsumer")
        self.deamon = True
        self._connection = connection
        self._consumer_tag = None
        self._router = connection._router
        self._exchange = connection._router._app_id
        self._queue = connection._router._router_id
        self._channel = None

    def run(self):
        """Start the Consumers.
        :return:
        """
        while True:
            try:
                
                
                channel = self._connection._connection.channel()
                self._channel = channel
                
                channel.exchange.declare(self._exchange, exchange_type="fanout", durable=False, auto_delete=True)
                
                channel.queue.declare(
                    self._queue,
                    auto_delete=True,
                    arguments = {
                        "x-message-ttl": 5000,
                        "x-max-length": 100
                    }
                )

                channel.queue.bind(self._queue, self._exchange, "ping")
                channel.queue.bind(self._queue, self._exchange, "session")
                channel.queue.bind(self._queue, self._exchange, "message")# "*." + self._router._connection_id + ".*")

                channel.basic.consume(self, self._queue, no_ack=True, no_local=False)
                channel.start_consuming()
                if not channel.consumer_tags:
                    channel.close()
            except amqpstorm.AMQPError as why:
                #LOGGER.exception(why)
                print("w", why)
                break
                #self.create_connection()
            except KeyboardInterrupt:
                #self.connection.close()
                break
        print("exit consumer")

    def stop(self):
        self._channel.stop_consuming()
        #self._channel.close()

    def __call__(self, message):
        self._connection._on_message(message)
        #message.ack()

class _PingThread(threading.Thread):
    def __init__(self, publisher, id):
        threading.Thread.__init__(self, None, None, "rbmqPing")
        self.deamon = True
        self._publisher = publisher
        self._router = publisher._connection._router
        self._interval = 5

        self._headers = {
            "connection_id": id,
            "routes": None,
            "topic" : "ping",
            "router_id": self._router._router_id
        }
        self._terminated = False

    def stop(self):
        self._terminated = True

    def run(self):
        while not self._terminated:
            routes = []
            for route in self._router.routes_in:
                routes.append(route.as_dict())
            json_routes = json.dumps(routes)
            properties = {
                'content_type': 'application/json',
                "headers": self._headers 
            }
            self._publisher._channel.basic.publish(
                json_routes,
                "ping",
                self._publisher._connection._app_id,
                properties
            )

            request_res = requests.post(
                "https://api.kervi.io/sessions",
                headers= {
                    "api-user": self._publisher._connection._user,
                    "api-password": self._publisher._connection._password,
                    "app_id": self._publisher._connection._app_id,
                    "app_name": self._publisher._connection._app_name
                } 
            )
            time.sleep(self._interval)

class _MQPublisher(object):
    def __init__(self, connection):
        self._connection = connection
        self._channel = self._connection._connection.channel()
        #self._channel.queue.declare('simple_queue')
        self._ping_thread = _PingThread(self, self._connection._router._connection_id)
        self._ping_thread.start()
        #self._bus_topic = None

    def stop(self):
        self._ping_thread.stop()
        self._channel.close()
    
    def send_message(self, topic, payload, headers={}):

        headers["topic"] = topic
        headers["router_id"] = self._connection._router._router_id
        properties = {
            'content_type': 'application/json',
            'headers': headers
        }
        json_body = json.dumps(payload, ensure_ascii=False)
        #print("sm",topic + self._connection._bus_topic, payload)
        self._channel.basic.publish(
            json_body,
            "message", #self._connection._bus_topic,
            self._connection._app_id,
            properties
        )

class _MQConnection(object):
    def __init__(self, router, user, password, address, port, api_key, app_id, app_name):
        #print("init mqc")
        self._max_retries = 10
        self._connection = None
        self._consumer = None
        self._publisher = None

        #self._bus_topic = ""
        self._consumer_tag = None
        self._user = user
        self._password = password
        self._address = address
        self._port = port
        self._vhost = api_key
        self._app_id = app_id
        self._app_name = app_name
        self._router = router
        self.connected = False


    def connect(self):
        """Create a connection.
        :return:
        """
        attempts = 0
        while True:
            attempts += 1
            try:
                if self._router._connection_type=="app":
                    request_res = requests.post(
                        "https://api.kervi.io/sessions",
                        headers= {
                            "api-user": self._user,
                            "api-password": self._password,
                            "app_id": self._app_id,
                            "app_name": self._app_name,
                            
                        }
                    )
                    print("cres", request_res.json())
                connection_string = 'amqps://'+self._user+':'+self._password+'@'+self._address+':' + str(self._port)+'/'+self._vhost
                #print("cns", connection_string)
                self._connection = UriConnection(
                    connection_string
                )

                #self._connection = Connection(self._address, self._user, self._password, port=self._port, vhost=self._vhost)
                break
            except amqpstorm.AMQPError as why:
                LOGGER.exception(why)
                print("why connect", why.error_code)
                if why.error_code == 403 or why.error_code == 530:
                    print("Kervi.io authentication error, check configuration for the plugin kervi.plugin.routing.kervi_io")
                    break

                if self._max_retries and attempts > self._max_retries:
                    break
                time.sleep(min(attempts * 2, 30))
            except KeyboardInterrupt:
                break
            
        if self._connection:
            
            self._consumer = _MQConsumer(self)
            self._consumer.start()
            self._publisher = _MQPublisher(self)
            self._router.on_connected()

    def disconnect(self):
        if self._consumer:
            self._consumer.stop()
        if self._publisher:
            self._publisher.stop()
        if self._connection:
            self._connection.close()

    def _on_message(self, message):
        
        if message.method["routing_key"] == "ping":
            self._router.on_ping(message.properties["headers"], message.json())
        elif message.method["routing_key"] == "session":
            self._router.on_session(message.properties["headers"], message.json())
        else:
            #print("Message:", message.json(), message.method, message.message_id, message.properties)
            self._router.on_message(message.properties["headers"], message.json())

    def send_message(self, topic, message, headers):
        if self._publisher:
            self._publisher.send_message(topic, message, headers)

    def register_bus_connection(self, connection):
        #self._bus_topic = ".Q." + "".join(self._router._connections.keys()) + ".Q"
        #print("bt", self._bus_topic)
        pass
        
    




class KerviBus:

    def __init__(self, config):
        pass

    @property
    def is_connected(self):
        raise NotImplementedError
    
    
    def reset(self, process_id, signal_port, ip="127.0.0.1", root_address=None, event_port=None):
        raise NotImplementedError

    def stop(self):
         raise NotImplementedError

    def send_command(self, command, *args, **kwargs):
        raise NotImplementedError

    def register_command_handler(self, command, func, **kwargs):
        raise NotImplementedError

    def unregister_command_handler(self, command, func, **kwargs):
        raise NotImplementedError

    def trigger_event(self, event, id, *args, **kwargs):
        raise NotImplementedError

    def register_event_handler(self, event, func, component_id=None, **kwargs):
        raise NotImplementedError
    
    def send_query(self, query, *args, **kwargs):
        raise NotImplementedError
    
    def register_query_handler(self, query, func, **kwargs):
        raise NotImplementedError

    def unregister_query_handler(self, query, func, **kwargs):
        raise NotImplementedError


import threading
import uuid
import datetime
import time
import json

import kervi.spine as spine
import kervi.utility.authorization_handler as authorization

class _PrepareThread(threading.Thread):
    def __init__(self, router):
        threading.Thread.__init__(self, "RouterPrepareThread")
        self.daemon = True
        self._router = router

    def run(self):
        self._router._prepare()


class _MQSession:
    def __init__(self, mq_session):
        self.mq_session = mq_session
        self.authenticated = False

class _Connection:
    def __init__(self, connection_id):
        self.connection_id = connection_id
        self.last_ping = datetime.datetime.now

class Route:
    def __init__(self, source_id, id, topic_type, topic):
        self.source_id = source_id
        self.id = id
        self.topic_type = topic_type
        self.topic = topic

    def as_dict(self):
        return {
            "source_id": self.source_id,
            "id": self.id,
            "topic_type": self.topic_type,
            "topic": self.topic
        }

class Router(object):
    def __init__(self, config):
        self._spine = spine.Spine()
        self._spine.register_event_handler("appReady", self._app_ready)

        self._spine.register_event_handler("moduleReady", self._module_ready)
        self._spine.register_event_handler("appReady", self._app_ready)
        self._routes_in = []
        self._routes_out = []
        self._ids = []
        self._prepare_thread = None
        self._route_table_ready = False
        self._router_id = uuid.uuid4().hex
        self._connection_id = None
        self._connections = {}
        self._mq_sessions= {}
        
    @property
    def routes_in(self):
        return self._routes_in

    @property
    def routes_out(self):
        return self._routes_out

    def _app_ready(self, id):
        self._connection_id = id
        self._prepare_thread = _PrepareThread(self)
        self._prepare_thread.start()

    def _module_ready(self, id):
        self._connection_id = id
        self._prepare_thread.start()

    def _read_routes(self, routes):
        if hasattr(routes, "__len__") and not isinstance(routes, dict):
            for route in routes:
                self._read_routes(route)
        else:
            for route in routes["routes"]:
                new_route = Route(routes["id"], route["id"], route["topic_type"], route["topic"])
                if route["direction"] == "in":
                    self._routes_in.append(new_route)
                else:
                    self._routes_out.append(new_route)

    def _prepare(self):
        self._internal_ids = []
        routes = self._spine.send_query("getComponentRoutes")
        self._read_routes(routes)
        self._spine._add_linked_handler(self._spine_handler)
        self._spine._add_linked_response_handler(self._on_response)     
        for route in self._routes_out:
            if route.topic_type == "event":
                self._spine.register_event_handler(route.topic, self._event_handler)
            if route.topic_type == "command":
                self._spine.register_command_handler(route.topic, self._command_handler)
        self._route_table_ready = True

    def _spine_handler(self, *args, **kwargs):
        topic = kwargs.get("topic_tag", None)
        
        if topic:
            if topic.startswith("query"):
                return self._query_handler(*args, **kwargs) 
            elif topic.startswith("command"):
                return self._command_handler(topic, *args, **kwargs)    

    def _event_handler(self, event, *args, **kwargs):
        topic = kwargs.pop("topic_tag", "event:")
        injected = kwargs.pop("injected", None)
        if injected != "KIO_" + self._router_id:
            message = {
                "args": args,
                "kwargs": kwargs
            }
            headers = {
                "type": "event",
                "topic": topic,
                "router_id": self._router_id
            }
            self.send_message(topic, message, headers)
    
    def _command_handler(self, command, *args, **kwargs):
        topic = kwargs.pop("topic_tag", "command:")
        injected = kwargs.pop("injected", None)
        if injected != "KIO_" + self._router_id:
            message = {
                "args": args,
                "kwargs": kwargs
            }
            headers = {
                "type": "command",
                "topic": topic,
                "router_id": self._router_id
            }
            self.send_message(topic, message, headers)
    
    def _query_handler(self, *args, **kwargs):
        topic = kwargs.pop("topic_tag", "query:")
        query_id = kwargs.pop("query_id", "0")
        response_address = kwargs.pop("response_address", None)
        injected = kwargs.pop("injected", None)
        if injected != "KIO_" + self._router_id:
            message = {
                "args": args,
                "kwargs": kwargs
            }
            headers = {
                "type": "query",
                "query_id": query_id,
                "response_address": response_address,
                "topic": topic,
                "router_id": self._router_id,
                "qts": str(time.time())
            }
            self.send_message(topic, message, headers)
            return "****no_response****"
        return None

    def start_router(self):
        raise NotImplementedError

    def stop(self):
        raise NotImplementedError

    def send_message(self, topic, payload, headers):
        raise NotImplementedError

    
    def on_ping(self, headers, routes):
        if headers["connection_id"] == self._connection_id:
            pass
        elif not headers["connection_id"] in self._connections:
            self._spine.log.verbose("new cloud connection:", headers["connection_id"])
            self._connections[headers["connection_id"]] = _Connection(headers["connection_id"])
            self.register_bus_connection(self._connections[headers["connection_id"]])
            for route in routes:
                if route["topic_type"] == "event":
                    self._spine.register_event_handler(route["topic"], self._event_handler)
                if route["topic_type"] == "command":
                    self._spine.register_command_handler(route["topic"], self._command_handler)
        else:
            self._connections[headers["connection_id"]].last_ping = datetime.datetime.now()

    def register_bus_connection(self, connection):
        
        pass
    
    def _on_response(self, event):
        self.send_message("query_response:", event["response"], event["headers"])
        
    def on_session(self, topic, headers, payload):
        #print("session", topic, self._router_id)
        if headers["router_id"] == self._router_id:

            if topic == "new":
                if headers["session_id"] in self._mq_sessions:
                    pass
                else:
                    session = _MQSession(headers["session_id"])
                    self._mq_sessions[headers["session_id"]] = session
                    
                    if authorization.active():
                        topic = "authenticate"
                    else:
                        session.authenticated = True
                        topic = "session_authenticated"
                        
                    self.send_message(topic, "", {})
        
                self._spine.log.verbose("new ui session", topic)
            
    def on_message(self, headers, payload):
        #print("om", self._router_id, headers, payload)
        topic = headers["topic"].split(":")
        if topic[0] == "session":
            self.on_session(topic[1], headers, payload)
            
        elif headers["router_id"] != self._router_id:
            #print("om", self._router_id, headers)
        
            if topic[0] == "event":
                #print("te", topic[1])
                event_id = None
                if len(topic) > 2:
                    event_id = topic[2]
                message_kwargs = dict(payload["kwargs"], injected="KIO_" + self._router_id)

                self._spine.trigger_event(topic[1], event_id, *payload["args"], **message_kwargs)
            elif topic[0] == "command":
                #print("c", topic, headers, payload)
                message_kwargs = dict(payload["kwargs"], injected="KIO_" + self._router_id)
                self._spine.send_command(topic[1], *payload["args"], **message_kwargs)
            elif topic[0] == "query":
                #print("oq", time.time()- float(headers["qts"]), topic, payload)
                headers = {
                    "type": "query_response",
                    "topic": "query_response:",
                    "response_address": headers["response_address"],
                    "query_id": headers["query_id"],
                    "router_id": self._router_id,
                    "qts": headers["qts"]
                }
                message_kwargs = dict(payload["kwargs"], injected="KIO_" + self._router_id, wait=False, headers=headers)
                res = self._spine.send_query(topic[1], *payload["args"], **message_kwargs)
                
            elif topic[0] == "query_response":
                self._spine.send_query_response(headers["response_address"],headers["query_id"], payload)
            elif topic[0] == "registerEventHandler":
                print("re", payload)
                self._spine.register_event_handler(payload["event"], self._event_handler, payload["eventId"])
        




class KerviPlugin(object):
    def __init__(self, name, config, manager):
        from kervi.config.configuration import _Configuration
        self._config = _Configuration()
        from kervi.config import Configuration
        self._global_config = Configuration
        plugin_config = {}
        if config:
            plugin_config = config.as_dict()
        self._config._load(config_user=plugin_config, config_base=self.get_default_config())
        
        self._name = name
        self._manager = manager

        from kervi.spine import Spine
        self.spine = Spine()
    
    @property
    def manager(self):
        return self._manager

    @property
    def plugin_config(self):
        return self._config

    @property
    def global_config(self):
        return self._global_config

    @property
    def name(self):
        return self._name

    def first_process_step(self):
        pass
    
    def process_step(self):
        pass

    def terminate_process(self):
        pass

    def get_default_config(self):
        return {}


import logging
import logging.config
import logging.handlers
import threading
from multiprocessing import Queue
_console_log = 15

class LogLevelFilter(logging.Filter):
    def __init__(self, param=None):
        self._level = param

    def filter(self, record):
        return record.levelno in self._level and record.levelno >= _console_log

class LogInfoLevelFilter(logging.Filter):
    def __init__(self, param=None):
        self._level = param

    def filter(self, record):
        if self._level == 25 and record.levelno == 25:
            return True

        if self._level == 25 and record.levelno == 10:
            return True
        
        if self._level == 10 and record.levelno ==10 :
            return True

        return False


def logger_thread(q):
    while True:
        record = q.get()
        if record is None:
            break
        logger = logging.getLogger(record.name)
        logger.handle(record)

class KerviLogHandler:
    def __init__(self, config):
        import platform
        
        log_level = config.log.level
            
        if log_level == "debug":
            log_level = logging.DEBUG
        
        if log_level == "verbose":
            log_level = 15

        if log_level == "info":
            log_level = logging.INFO
        
        if log_level == "warning":
            log_level = logging.WARNING

        if log_level == "error":
            log_level = logging.ERROR

        console_level = config.log.console_level
        
        if console_level == "debug":
            _console_log = logging.DEBUG
        
        if console_level == "verbose":
            _console_log = 15

        if console_level == "info":
            _console_log = logging.INFO
        
        if log_level == "warning":
            _console_log = logging.WARNING

        if log_level == "error":
            _console_log = logging.ERROR

        log_config = {
            'version': 1,
            'filters':{
                'info':{
                    '()': LogLevelFilter,
                    'param': [logging.INFO]
                },
                'verbose':{
                    '()': LogLevelFilter,
                    'param': [15]
                },
                'warning':{
                    '()': LogLevelFilter,
                    'param': [logging.WARNING]
                },
                'error':{
                    '()': LogLevelFilter,
                    'param': [logging.ERROR]
                }
            },
            'formatters': {
                'detailed': {
                    'class': 'logging.Formatter',
                    'format': '%(asctime)s %(name)-15s %(levelname)-8s %(processName)-10s %(message)s'
                },
                'console': {
                    'class': 'logging.Formatter',
                    'format': '\033[92m%(message)s \033[0m',
                },
                'console-verbose': {
                    'class': 'logging.Formatter',
                    'format': '\33[90m   %(message)s \33[0m',
                },
                'console-warning': {
                    'class': 'logging.Formatter',
                    'format': '\33[93m%(message)s \33[0m',
                },
                'console-error': {
                    'class': 'logging.Formatter',
                    'format': '\033[91m%(levelname)-8s %(processName)-10s %(message)s \033[0m',
                },
            },
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'level': 'INFO',
                    'formatter': 'console',
                    'filters': ['info']
                },
                'console-verbose': {
                    'class': 'logging.StreamHandler',
                    'level': 15,
                    'formatter': 'console-verbose',
                    'filters': ['verbose']
                },
                'console-warning': {
                    'class': 'logging.StreamHandler',
                    'level': 'WARNING',
                    'formatter': 'console-warning',
                    'filters': ['warning']
                },
                'console-error': {
                    'class': 'logging.StreamHandler',
                    'level': 'ERROR',
                    'formatter': 'console-error',
                    'filters': ['error']
                },
                'file': {
                    'class': 'logging.handlers.RotatingFileHandler',
                    'filename': config.log.file,
                    'mode': 'w',
                    'backupCount': 3,
                    'maxBytes': config.log.max_file_size,
                    'formatter': 'detailed',
                },
            },
            'loggers':{
                'kervi':{
                    'handlers': ['console', 'file']
                }
            },
            'root': {
                'level': log_level,
                'handlers': ['console', 'console-verbose', 'console-warning', 'console-error', 'file']
            },
        }
        logging.config.dictConfig(log_config)
        self._log_queue = Queue()
        self._logging_thread = threading.Thread(target=logger_thread, args=(self._log_queue,))
        self._logging_thread.deamon = True
        if platform.system() == "Windows":
            self._logging_thread.start()
        #return KerviLog("application")

    def stop(self):
        self._log_queue.put_nowait(None)


#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

import os
import json
from kervi.plugin.routing.routing_plugin import RouterPlugin
from kervi.plugin.routing.kervi_io.rbmq import _MQConnection

            
class KerviIORouterPlugin(RouterPlugin):
    def __init__(self, config, manager):
        RouterPlugin.__init__(self, "KERVI_IO", config, manager)
        self._router_name = "kervi.io"
        self._address = self.plugin_config.address
        self._port = self.plugin_config.port
        self._app_id = self.global_config.application.id
        if os.path.isfile('api_credentials.kervi_io'):
            with open('api_credentials.kervi_io') as handle:
                self.spine.log.verbose("using credentials file")
                json_data = json.loads(handle.read())
                self._api_user = json_data["api_user"]
                self._api_password = json_data["api_password"]
                self._api_key = json_data["api_channel"]
        else:
            self._api_user = self.plugin_config.api_user
            self._api_password = self.plugin_config.api_password
            self._api_key = self._config.api_channel
        
        self._app_name = self.global_config.application.get("name", None)
        self._enabled = self._config.enabled
        self._connection = _MQConnection(
            self,
            self._api_user,
            self._api_password,
            self._address,
            self._port,
            self._api_key,
            self._app_id,
            self._app_name
        )
        
    def get_default_config(self):
        return{
            "enabled": False,
            "address": "mq.kervi.io",
            "port": 5671,
            "api_user": "",
            "api_password": "",
            "api_channel": ""
        }
    
    def connect(self):
        pass    
        #self._thread = _IncomingThread(self._channel_in)

    def start_router(self):
        self._connection.connect()

    def stop_router(self):
        self._connected = False
        self._connection.disconnect()

    def register_route(self, route):
        pass
    
    def register_bus_connection(self, connection):
        self._connection.register_bus_connection(connection)
    
    def send_message(self, topic, payload, headers):
        if self._connected:
            self._connection.send_message(topic, payload, headers)



from kervi.plugin.message_bus.kervi_bus_plugin import KerviBusPlugin
from kervi.plugin.plugin_manager import PluginManager
from kervi.config import Configuration
from kervi.core.utility.kervi_logging import KerviLog

class BusManager():
    def __init__(self, log_queue=None):
        
        self._plugin_manager = None
        self._plugin_manager = PluginManager(Configuration, "message_bus", [KerviBusPlugin], load_silent=True, log_queue=log_queue)
        self._plugin_manager.load_managed_plugins()
        self._current_bus = None
        self._log = KerviLog("BusManager")

    def load(self, process_id, spine_port, root_address = None, ip="127.0.0.1"):
        self._plugin_manager.load_managed_plugins()
        
        for plugin in self._plugin_manager.plugins:
            plugin.load(process_id, spine_port, root_address, ip)
            self._current_bus = plugin.bus
            break
        else:
            self._log.error("No message bus plugin found")

        if self._current_bus:
            import kervi.spine as spine
            spine.set_spine(self._current_bus)

    def stop(self):
        self._current_bus.stop()
    
    @property
    def bus(self):
        return self._current_bus


#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

def init_plugin(config, manager):
    from kervi.plugin.webserver.build_in.web_server import KerviWebServerPlugin
    return KerviWebServerPlugin("KerviWebServer", config, manager)

def plugin_type():
    return "webserver"


#Copyright 2017 Tim Wentlau.
#Distributed under the MIT License. See LICENSE in root of project.

def init_plugin(config, manager):
    from kervi.plugin.routing.kervi_io.mq_router import KerviIORouterPlugin
    return KerviIORouterPlugin(config, manager)

def plugin_type():
    return "routing"